#include "cps.h"
// CPS - Run

// Inputs:
unsigned char CpsReset=0;
unsigned char Cpi01A=0,Cpi01C=0,Cpi01E=0;

static int nInterrupt;
static int nIrqCycles;

static int DrvReset()
{
  // Reset machine
  CpsEepromReset();

  SekOpen(0); SekReset();
  SekClose();

  ZetReset(); // Reset z80

  if(Cps==2) {
    // Disable beam-synchronized interrupts
    *((unsigned short*)(CpsReg + 0x4E)) = 0x0200;
    *((unsigned short*)(CpsReg + 0x50)) = 0x0106;
    *((unsigned short*)(CpsReg + 0x52)) = 0x0106;
  }

  return 0;
}

int CpsRunInit()
{
  int nRet=0;
  SekInit(1); // Allocate 68000

  // Memory init
  nRet=CpsMemInit(); if (nRet!=0) return 1;
  CpsEepromBlank();
  CpsRwInit();

  // Sprite Masking init
  if(ZBuf!=NULL) nMaskAct=1;

  // Graphics init
  nRet=CpsPalInit(); if (nRet!=0) return 1;
  nRet=CpsObjInit(); if (nRet!=0) return 1;

  // Sound init
  if (Cps==1 && Cps1Qs==0) { nRet=PsndInit(); if (nRet!=0) return 1; }
  if (Cps==2 || Cps1Qs==1) { nRet=QsndInit(); if (nRet!=0) return 1; }

  CpsEepromReset();
  DrvReset();

  //Init Draw Function
  DrawFnInit();

  nStateNVMin = 0x010002;
  nStateAllMin = 0x008400;
  return 0;
}

int CpsRunExit()
{
  // Sound exit
  if (Cps==2 || Cps1Qs==1) QsndExit();
  if (Cps==1 && Cps1Qs==0) PsndExit();

  // Graphics exit
  CpsObjExit();
  CpsPalExit();

  //Sprite Masking exit
  nMaskAct=0;
  ZBuf=NULL;

  // Memory exit
  CpsRwExit();
  CpsMemExit();

  SekExit(); // Deallocate 68000

  return 0;
}

// nStart = 0-3, nCount=1-4
static INLINE void GetPalette(int nStart,int nCount)
{
  int nPal; unsigned char *Find;
  // Update Palette
  nPal=*((unsigned short *)(CpsReg+0x0a)); nPal<<=8;

  // Ghouls points to the wrong place on boot up I think:
  nPal&=0xfff800; Find=CpsFindGfxRam(nPal,0x1000);
  if (Find==NULL) return;
  memcpy(CpsSavePal+(nStart<<10), Find+(nStart<<10), nCount<<10);
}

static INLINE void CopyCpsReg(int i)
{
	memcpy(CpsSaveReg[i], CpsReg, 0x0100);
}

static INLINE void CopyCpsFrg(int i)
{
	memcpy(CpsSaveFrg[i], CpsFrg, 0x0010);
}

// Schedule a beam-synchronized interrupt. Returns the number of cycles to the next interrupt (must be > 0).
static void ScheduleIRQ()
{
	int nNextIrq = 0;
	if (*((unsigned short*)(CpsReg + 0x50)) < 0x0106 && ~*((unsigned short*)(CpsReg + 0x4E)) & 0x0200) {
		nNextIrq = (*((unsigned short*)(CpsReg + 0x50)) * nCpsCycles / 0x0106) + 1;
	}
	else if (*((unsigned short*)(CpsReg + 0x52)) < 0x0106 && ~*((unsigned short*)(CpsReg + 0x4E)) & 0x0200) {
		nNextIrq = (*((unsigned short*)(CpsReg + 0x52)) * nCpsCycles / 0x0106) + 1;
	}
	else {
		nIrqCycles = nCpsCycles + 1;
		return;
	}

	if (nNextIrq > nIrqCycles) {
		nIrqCycles = nNextIrq;
	}
	else {
		nIrqCycles = nCpsCycles + 1;
	}
	return;
}

static void DoIRQ(int *nDone)
{
	int nLineReg;

	// 0x4E - bit 9 = 1: Beam Synchronized interrupt #1 disabled
	// 0x50 - Beam synchronized interrupt #1 occurs at raster line.
	// 0x52 - Beam synchronized interrupt #2 occurs at raster line.

	// Check for beam synchronized interrupt (in case the CPU modified registers).
	if (*((unsigned short*)(CpsReg + 0x50)) < 0x0106 && ~*((unsigned short*)(CpsReg + 0x4E)) & 0x0200) {
		nLineReg = 0x50;
	}
	else if (*((unsigned short*)(CpsReg + 0x52)) < 0x0106 && ~*((unsigned short*)(CpsReg + 0x4E)) & 0x0200) {
		nLineReg = 0x52;
	}
	else {
		return;
	}

	// Trigger IRQ and copy registers.
	if (*((unsigned short*)(CpsReg + nLineReg)) >= 0x08) {
		int nCurrentline = *((unsigned short*)(CpsSaveReg[nInterrupt] + nLineReg)) - 0x08;

 		// If this interrupt occurs very near to the previous one, ignore the previous one.
		if (nCurrentline >= nRasterline[nInterrupt] + 0x10) {
			nInterrupt++;
			nRasterline[nInterrupt] = nCurrentline;
		}
	}
	SekInterrupt(4);
	*nDone += SekRun(nCpsCycles * 0x01 / 0x0106);
	if (nRasterline[nInterrupt] < 224) {
		CopyCpsReg(nInterrupt);
		CopyCpsFrg(nInterrupt);
	} else {
		nRasterline[nInterrupt] = 0;
	}
	return;
}

int Cps1Frame()
{
  int nDone,nTotal,nNext,i;

  if (CpsReset) DrvReset();

  CpsRwGetInp(); // Update the input port values

  nDone=0; nTotal=nCpsCycles;

  SekOpen(0);

  SekInterrupt(2); // VBlank
  for (i=0; i<4; i++)
  {
    int nStart,nEnd;
    nNext=((i+1)*nTotal) >> 2; // find out next cycle count to run to
    nDone+=SekRun(nNext-nDone); // run cpu

    CpsScodePoll(); // See which code has been passed
    // Run sound chips for this section
    nStart=i<<8; nEnd=(i+1)<<8;

	if (Cps1Qs==1) QsndSectRun(nStart,nEnd);
    else           PsndSectRun(nStart,nEnd);
  }
  SekClose();

  GetPalette(0,4); // Get palette
  CpsObjGet();     // Get objects
  memcpy(CpsSaveReg[0],CpsReg,0x100); // Registers correct now

  if (pBurnDraw==NULL) return 0; // Done

  // Draw the screen -----------------
  CpsDraw();

	extern int counter;
	if (counter) {
		printf("\n");
		for (i = 0; i < 0x080; i+= 8) {
			printf("%2x: %4x %4x %4x %4x %4x %4x %4x %4x\n", i * 2, *((unsigned short*)(CpsSaveReg[nInterrupt] + 0 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 2 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 4 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 6 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 8 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 10 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 12 + i * 2)), *((unsigned short*)(CpsSaveReg[nInterrupt] + 14 + i * 2)));
		}
		extern int bRunPause;
		bRunPause = 1;
		counter = 0;
	}

  return 0;
}

int Cps2Frame()
{
	int nDone,nTotal,nNext,i;

	if (CpsReset) {
		DrvReset();
	}

	CpsRwGetInp(); 										// Update the input port values

	nTotal = nCpsCycles - (nCpsCycles * 0x1E / 0x0106);	// Account for VBlank
	nDone = 0;

	nInterrupt = 0;
	nIrqCycles = 0;
	for (i = 0; i < MAX_RASTER + 2; i++) {
		nRasterline[i] = 0;
	}

	SekOpen(0);
	ScheduleIRQ();

//	nDone += SekRun(nCpsCycles * 0x08 / 0x0106);

	CopyCpsReg(0);
	CopyCpsFrg(0);

	for (i = 0; i < 4; i++) {
		int nStart,nEnd;

		nNext = ((i + 1) * nTotal) >> 2;			// find out next cycle count to run to
		while (nNext > nIrqCycles && nInterrupt < MAX_RASTER) {
			nDone += SekRun(nIrqCycles - nDone);
			DoIRQ(&nDone);
			ScheduleIRQ();
		}
		nDone += SekRun(nNext - nDone);				// run cpu

		CpsScodePoll();								// See which code has been passed

		// Run sound chips for this section
		nStart=i<<8; nEnd=(i+1)<<8;
		QsndSectRun(nStart,nEnd);
	}

	GetPalette(0, 4);								// Get palette

	if (pBurnDraw) {
		CpsDraw();
	}

	CpsObjGet();									// Get objects

	SekInterrupt(2);								// VBlank
	nDone += SekRun(nCpsCycles - nTotal);
	SekClose();

	return 0;
}

